import numpy as np
from sklearn.decomposition import PCA

RANDOM_STATE = 42


def get_pca_statistics(data: np.ndarray) -> np.ndarray:
    """
    Считает долю дисперсии для каждого компонента.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).

    Returns
    -------
    np.ndarray
        Массив доль дисперсии, размера N.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    pca = PCA(n_components=data.shape[1], random_state=RANDOM_STATE)
    pca.fit(data)
    return pca.explained_variance_ratio_


def pca_transform(data: np.ndarray, n_components: int) -> np.ndarray:
    """
    Понижение размерности с помощью PCA путём проецирования данных на главные (собственные) вектора.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).
    n_components: int
        Количество сохраняемых компонентов.

    Returns
    -------
    np.ndarray
        Преобразованный массив, размера (N, n_components).
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert (
        n_components > 0
    ), "Параметр n_components должен принимать положительное значение"

    return PCA(n_components=n_components, random_state=RANDOM_STATE).fit_transform(data)



import numpy as np

from .curve_utils import get_peaks, remove_given_width_peaks


def fill_with_value_by_mask(
    data: np.ndarray, mask: np.ndarray, blank_code: float
) -> np.ndarray:
    """
    Востанавливает массив, заполняя кодом бланковки по маске.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером N.
    mask: np.ndarray
        Маска(True - в строке был nan, False - в строке не было nan), размером M >= N.
    blank_code: float
        Код бланковки.

    Returns
    -------
    np.ndarray
        Востановленный массив, размером M.
    """

    assert data.ndim == 1, "data: Ожидался 1D массив"
    assert mask.ndim == 1, "mask: Ожидался 1D массив"
    assert (
        mask.shape[0] >= data.shape[0]
    ), "Длина массива mask должна быть не меньше длины data"

    data_with_blank_code = np.zeros(mask.shape[0])
    data_with_blank_code[mask] = blank_code
    data_with_blank_code[~mask] = data
    return data_with_blank_code


def curve_peaks_processing(
    curve: np.ndarray, depth: np.ndarray, minimal_width_in_meter: float
) -> np.ndarray:
    """
    Обработка пиков кривой. Удаление пиков, которые меньше заданной ширины (minimal_width_in_meter).

    Parameters
    ----------
    curve: np.ndarray
        Массив значений кривой, размером N.
    depth: np.ndarray
        Массив глубин кривой, размером N.
    minimal_width_in_meter: float
        Минимальная ширина пиков в метрах.

    Returns
    -------
    np.ndarray
        Обработанная кривая, размером N.
    """

    assert curve.ndim == 1, "curve: Ожидался 1D массив"
    assert depth.ndim == 1, "depth: Ожидался 1D массив"
    assert (
        minimal_width_in_meter > 0
    ), "Параметр minimal_width_in_meter должен принимать положительное значение"

    dz = np.diff(depth)[0]
    minimal_width_in_sample = int(minimal_width_in_meter / dz)

    given_width_peaks = get_peaks(curve, minimal_width_in_sample)
    filled_curve = remove_given_width_peaks(
        curve, given_width_peaks, minimal_width_in_sample
    )
    if np.unique(filled_curve).size == 2:
        filled_curve = abs(filled_curve - 1)
    return filled_curve



from typing import List, Optional

import numpy as np
from cratonapi import datacontainers
from cratonapi.dataconnector import DataConnector
from scipy.interpolate import RegularGridInterpolator

from cratonml.exceptions.DataExceptions import EmptyInputDataException


class Grid:
    """
    Класс для работы с гридами.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(
        self, grid_id_list: list[int], grid_type_list: list[int]
    ) -> list[datacontainers.Grid]:
        """
        Подгружает cетки(гриды).

        Parameters
        ----------
        grid_id_list: list[int]
            Список идентификаторов гридов.
        grid_type_list: list[int]
            Список типов гридов. Тип грида: 0 — обычный, 1 — смешанный.

        Returns
        -------
        list[cratonapi.datacontainers.Grid]
            Список гридов.
        """

        assert len(grid_id_list) == len(
            grid_type_list
        ), "grid_id_list и grid_type_list должны иметь одинаковую длину"
        if len(grid_id_list) == 0:
            raise EmptyInputDataException(
                "Список идентификаторов и список типов гридов пустые"
            )

        grid_list = []
        for grid_id, grid_type in zip(grid_id_list, grid_type_list):
            if grid_type == 0:
                grid = self.connection.get_grid(grid_id)
                if grid.data.shape[0] != 0:
                    grid_list.append(grid)
            elif grid_type == 1:
                grids = self.connection.get_grid_sp(grid_id)
                for grid in grids:
                    grid_list.append(grid)
        return grid_list

    def save(self, name: str, grid: datacontainers.Grid) -> None:
        """
        Загружает грид в Desmana.

        Parameters
        ----------
        name: str
            Имя грида.
        grid: cratonapi.datacontainers.Grid
            Сеточная модель.
        """

        self.connection.upload_grid(name, grid)

    def get_info(self) -> dict:
        """
        Подгружает информацию о гридах.

        Returns
        -------
        dict
            Словарь, где ключ - имя грида.
            Значения:
                id: int
                    Идентификатор грида.
                type: int
                    Тип грида: 0 — обычный, 1 — смешанный.
                visibility: int
                    Видимость грида: 0 — скрытый на планшете, 1 — отображается на планшете.
        """

        grids_list = self.connection.get_grids_list()
        info = {
            grid.grid_name: {
                "id": grid.grid_id,
                "type": grid.grid_type,
                "visibility": grid.grid_visibility,
            }
            for grid in grids_list
        }
        return info

    @staticmethod
    def numpy_to_grid(
        data: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
        blank_code: float,
        grid_id: int = 0,
    ) -> datacontainers.Grid:
        """
        Преобразовывает numpy массив в грид.

        Parameters
        ----------
        data: np.ndarray
            1D массив значений грида.
        x: np.ndarray
            Сетка координат по x в виде 2D массива.
        y: np.ndarray
            Сетка координат по y в виде 2D массива.
        blank_code: float
            Код бланковки.
        grid_id: int, default=0
            Идентификатор грида.

        Returns
        -------
        cratonapi.datacontainers.Grid
            Грид.
        """

        assert data.ndim == 1, "data: Ожидался 1D массив"
        assert x.ndim == 2, "x: Ожидался 2D массив"
        assert y.ndim == 2, "y: Ожидался 2D массив"
        assert grid_id >= 0, "grid_id должен принимать неотрицательное значение"

        data_to_save = np.nan_to_num(data, nan=blank_code)
        grid_to_save = datacontainers.Grid(
            n_id=grid_id,
            n_x=x.shape[1],
            n_y=x.shape[0],
            x_min=np.min(x),
            x_max=np.max(x),
            y_min=np.min(y),
            y_max=np.max(y),
            z_min=np.nanmin(data[data != blank_code]),
            z_max=np.nanmax(data[data != blank_code]),
            blank_code=blank_code,
            data=data_to_save,
        )
        return grid_to_save

    @staticmethod
    def parse_to_numpy(
        grid_list: list[datacontainers.Grid],
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Преобразовывает список гридов в numpy массив.

        Parameters
        ----------
        grid_list: list[cratonapi.datacontainers.Grid]
            Список гридов.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            2D массив значений грида.
            Сетка координат по x в виде 2D массива.
            Сетка координат по y в виде 2D массива.
        """

        x_min = np.max([grid.x_min for grid in grid_list])
        x_max = np.min([grid.x_max for grid in grid_list])

        y_min = np.max([grid.y_min for grid in grid_list])
        y_max = np.min([grid.y_max for grid in grid_list])

        dx = np.min([(grid.x_max - grid.x_min) / grid.n_x for grid in grid_list])
        dy = np.min([(grid.y_max - grid.y_min) / grid.n_y for grid in grid_list])

        x = np.arange(x_min, x_max, dx)
        y = np.arange(y_min, y_max, dy)

        xx, yy = np.meshgrid(x, y)
        grid_data_list: List[np.ndarray] = []
        mask_list: List[np.ndarray] = []
        for grid in grid_list:
            blank_code = grid.blank_code
            x_grid = np.linspace(grid.x_min, grid.x_max, grid.n_x)
            y_grid = np.linspace(grid.y_min, grid.y_max, grid.n_y)
            grid_2d = np.reshape(grid.data, (grid.n_y, grid.n_x))
            grid_2d[grid_2d == blank_code] = np.nan

            interp_func = RegularGridInterpolator((y_grid, x_grid), grid_2d)
            new_grid = interp_func((yy, xx))
            mask = np.isnan(new_grid)
            mask_list.append(mask.ravel())
            grid_data_list.append(new_grid.ravel())

        mask_array: np.ndarray = np.asarray(mask_list)
        general_mask = np.logical_or.reduce(mask_array).T
        data: np.ndarray = np.asarray(grid_data_list).T
        data[general_mask] = np.nan
        return data, xx, yy

    @staticmethod
    def get_attribute_values_by_coords(
        grid_data: np.ndarray,
        grid_xx: np.ndarray,
        grid_yy: np.ndarray,
        coords: np.ndarray,
    ) -> Optional[np.ndarray]:
        """
        Находит значение гридов в точке.

        Parameters
        ----------
        grid_data: np.ndarray
            Массив значений гридов размера (X*Y, M) или (X*Y,)
        grid_xx: np.ndarray
            Сетка координат по x в виде 2D массива размера (X, Y).
        grid_yy: np.ndarray
            Сетка координат по y в виде 2D массива размера (X, Y).
        coords: np.ndarray
            Координаты точки в виде 1D массива длины 2.

        Returns
        -------
        Optional[np.ndarray]
            Массив длины M значений гридов в точке.
        """

        assert (
            grid_data.ndim == 1 or grid_data.ndim == 2
        ), "grid_data: Ожидался 1D или 2D массив"
        assert grid_xx.ndim == 2, "grid_xx: Ожидался 2D массив"
        assert (
            grid_xx.shape == grid_yy.shape
        ), "grid_xx, grid_yy: Ожидались массивы одинакового размера"
        assert (
            coords.ndim == 1 and len(coords) == 2
        ), "coords: Ожидался 1D массив длины 2"
        assert (
            grid_data.shape[0] == grid_xx.shape[0] * grid_xx.shape[1]
        ), "grid_data: Ожидался массив длины grid_xx.shape[0] * grid_xx.shape[1]"

        if np.any(np.isnan(coords)):
            if grid_data.ndim == 1:
                return np.tile(np.nan, 1)
            if grid_data.ndim == 2:
                return np.tile(np.nan, grid_data.shape[1])
        else:
            x_idx = np.argmin(np.abs(grid_xx - coords[0])[0, :])
            y_idx = np.argmin(np.abs(grid_yy - coords[1])[:, 0])
            if grid_data.ndim == 1:
                return grid_data[y_idx * grid_xx.shape[1] + x_idx]
            if grid_data.ndim == 2:
                return grid_data[y_idx * grid_xx.shape[1] + x_idx, :]

        return None



from typing import Tuple

import numpy as np
from sklearn.inspection import permutation_importance
from sklearn.metrics import precision_recall_fscore_support
from sklearn.model_selection import KFold, cross_val_score

RANDOM_STATE = 42


def calculate_metrics(
    y_true: np.ndarray, y_pred: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    precision, recall, fbeta_score, support = precision_recall_fscore_support(
        y_true, y_pred
    )
    return precision, recall, fbeta_score, support


def k_fold(
    y_train: np.ndarray, x_train: np.ndarray, model: object, n_epochs: int = 10
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    kf = KFold(n_splits=min(len(y_train), n_epochs))
    if x_train.ndim == 1:
        x_train = x_train.reshape(-1, 1)
    precision = cross_val_score(
        model, x_train, y_train.reshape(-1, 1), cv=kf, scoring="precision_weighted"
    )
    recall = cross_val_score(
        model, x_train, y_train.reshape(-1, 1), cv=kf, scoring="recall_weighted"
    )
    accuracy = cross_val_score(
        model, x_train, y_train.reshape(-1, 1), cv=kf, scoring="accuracy"
    )
    f1 = cross_val_score(
        model, x_train, y_train.reshape(-1, 1), cv=kf, scoring="f1_weighted"
    )
    return precision, recall, accuracy, f1


def feature_importanse(
    model: object, x_test: np.ndarray, y_test: np.ndarray
) -> np.ndarray:
    if hasattr(model, "feature_importances_"):
        importances = model.feature_importances_
    else:
        results = permutation_importance(
            model, x_test, y_test, n_repeats=30, random_state=RANDOM_STATE
        )
        importances = results.importances_mean
    return importances



import numpy as np
from sklearn.ensemble import RandomForestClassifier as RandomForest
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier as DecisionTree
from typing_extensions import Optional


class GaussianNaiveBayesClassifier:
    """Класс для классификации алгоритмом Gaussian Naive Bayes."""

    @staticmethod
    def get_model(x_train: np.ndarray, y_train: np.ndarray) -> GaussianNB:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.
        y_train: np.ndarray
            Массив значений размера N.

        Returns
        -------
        GaussianNB
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert y_train.ndim == 1, "y_train: Ожидался 1D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Массив y_train имеет несоответствующую длину"

        model = GaussianNB()
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: GaussianNB, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует метки для тестового набора данных.

        Parameters
        ----------
        model: GaussianNB
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction


class LogisticRegressionClassifier:
    """Класс для классификации алгоритмом Logistic Regression."""

    @staticmethod
    def get_model(x_train: np.ndarray, y_train: np.ndarray) -> LogisticRegression:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.
        y_train: np.ndarray
            Массив значений размера N.

        Returns
        -------
        LogisticRegression
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert y_train.ndim == 1, "y_train: Ожидался 1D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Массив y_train имеет несоответствующую длину"

        model = LogisticRegression()
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: LogisticRegression, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует метки для тестового набора данных.

        Parameters
        ----------
        model: LogisticRegression
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction


class DecisionTreeClassifier:
    """Класс для классификации Decision Tree."""

    @staticmethod
    def get_model(
        x_train: np.ndarray,
        y_train: np.ndarray,
        min_samples_split: Optional[int] = 2,
        max_depth: Optional[int] = None,
    ) -> DecisionTree:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.
        y_train: np.ndarray
            Массив значений размера N.
        min_samples_split: int or float = 2
            Минимальное количество выборок, необходимое для разделения внутреннего узла.(параметр DecisionTree).
        max_depth: int = None
            Максимальная глубина дерева(параметр DecisionTree).

        Returns
        -------
        DecisionTree
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert y_train.ndim == 1, "y_train: Ожидался 1D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Массив y_train имеет несоответствующую длину"

        model = DecisionTree(min_samples_split=min_samples_split, max_depth=max_depth)
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: DecisionTree, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует метки для тестового набора данных.

        Parameters
        ----------
        model: DecisionTree
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction


class RandomForestClassifier:
    """Класс для классификации Random Forest."""

    @staticmethod
    def get_model(
        x_train: np.ndarray,
        y_train: np.ndarray,
        n_estimators: int = 100,
        max_depth: Optional[int] = None,
    ) -> RandomForest:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.
        y_train: np.ndarray
            Массив значений размера N.
        n_estimators: int = 100
            Количество деревьев в лесу.(параметр RandomForest).
        max_depth: int = None
            Максимальная глубина дерева(параметр RandomForest).
        Returns
        -------
        RandomForest
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert y_train.ndim == 1, "y_train: Ожидался 1D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Массив y_train имеет несоответствующую длину"

        model = RandomForest(n_estimators=n_estimators, max_depth=max_depth)
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: RandomForest, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует метки для тестового набора данных.

        Parameters
        ----------
        model: RandomForest
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (N, M). Где N - количество точек, M - количество атрибутов.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction



"""Модуль исключений при работе с данными(гриды, кубы, скважины, контуры)."""


class DataException(BaseException):
    """Базовый класс исключений при работе с данными(гриды, кубы, скважины, контуры)."""

    pass


class EmptyInputDataException(DataException):
    """"""

    pass



import numpy as np
from scipy.signal import find_peaks


def remove_given_width_peaks(
    data: np.ndarray, peaks: np.ndarray, width: int
) -> np.ndarray:
    """Удаляет из массива data пики по индексам peaks с разрешенной минимальной шириной width."""

    window_size = width + width // 2 + width % 2
    filled_data = np.copy(data)
    for i in peaks:
        peak_left = 0
        peak_right = len(data)
        for j in range(1, width + 1):
            if i - j >= 0 and peak_left == 0:
                if data[i] != data[i - j]:
                    peak_left = i - j + 1
            if i + j < len(data) and peak_right == len(data):
                if data[i] != data[i + j]:
                    peak_right = i + j
        left = max(0, i - window_size)
        right = min(len(data) - 1, i + window_size + 1)

        unique, counts = np.unique(filled_data[left:right], return_counts=True)
        val = unique[np.where(counts == max(counts))[0][0]]
        filled_data[peak_left:peak_right] = val
    return filled_data.astype("int")


def get_peaks(curve: np.ndarray, width: int) -> np.ndarray:
    """Находит пики в кривой curve с шириной не более width. Возвращает индексы пиков."""

    peaks = []
    for i in range(np.max(curve) + 1):
        data = np.zeros(curve.shape)
        mask = curve == i
        data[mask] = 1
        new_peaks, _ = find_peaks(data, width=[0, width])
        if i == 0:
            peaks = new_peaks
        peaks = np.unique(np.hstack((new_peaks, peaks)))

    for i in range(1, width):
        if curve[0] != curve[i]:
            peaks = np.unique(np.hstack(([i // 2], peaks)))
            break
    for i in range(1, width):
        if curve[len(curve) - 1] != curve[len(curve) - 1 - i]:
            peaks = np.unique(np.hstack(([len(curve) - 1 - i // 2], peaks)))
            break

    peaks_array = np.asarray(peaks)
    return peaks_array



from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
from cratonapi import datacontainers
from cratonapi.dataconnector import DataConnector
from cratonapi.datacontainers import Curve
from scipy.interpolate import interp1d

from cratonml.exceptions.DataExceptions import EmptyInputDataException
from cratonml.exceptions.WellExceptions import (
    WellEmptyCurvesException,
    WellTrajectoryDepthException,
)


class Well:
    """
    Класс для работы со скважинами.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(
        self, well_id: int, well_curves_name: list[str]
    ) -> list[datacontainers.WellCurve]:
        """
        Подгружает видимые кривые из скважины.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.
        well_curves_name: list[str]
            Список имен кривых.

        Returns
        -------
        list[cratonapi.datacontainers.WellCurve]
            Список кривых для скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"
        if len(well_curves_name) == 0:
            raise EmptyInputDataException("Список имен кривых пустой")

        well_curves = self.connection.get_well_curves(well_id)
        curves = []
        for curve in well_curves:
            if curve.curve_name in well_curves_name:
                curves.append(curve)
        return curves

    def read_all_curves(
        self, well_id: int, well_curves_name: list[str], well_curves_id: list[int]
    ) -> list[datacontainers.WellCurve | datacontainers.Curve]:
        """
        Подгружает кривые из скважины.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.
        well_curves_name: list[str]
            Список имен кривых.
        well_curves_id: list[int]
            Список идентификаторов кривых.

        Returns
        -------
        list[cratonapi.datacontainers.WellCurve | cratonapi.datacontainers.Curve]
            Список кривых для скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"
        assert len(well_curves_name) == len(
            well_curves_id
        ), "well_curves_name и well_curves_id должны иметь одинаковую длину"

        if len(well_curves_name) == 0:
            raise EmptyInputDataException(
                "Список имен и список идентификаторов кривых пустые"
            )

        curves_name = well_curves_name.copy()
        curves_id = well_curves_id.copy()

        well_curves = self.connection.get_well_curves(well_id)

        curves = []
        for curve in well_curves:
            if curve.curve_name in curves_name:
                curves.append(curve)
                idx = curves_name.index(curve.curve_name)
                curves_name.remove(curve.curve_name)
                curves_id.pop(idx)

        for curve_id in curves_id:
            if curve_id is None:
                curves.append(
                    Curve(curve_id=None, point_values=[np.nan], point_depths=[np.nan])
                )
            else:
                assert (
                    curve_id >= 0
                ), "curves_id должен принимать неотрицательные значения"
                curves.append(self.read_curve(well_id, curve_id))
        return curves

    def read_curve(self, well_id: int, curve_id: int) -> datacontainers.Curve:
        """
        Подгружает кривую из скважины.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.
        curve_id: int
            Идентификатор кривой.

        Returns
        -------
        cratonapi.datacontainers.Curve
            Список кривых для скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"
        assert curve_id >= 0, "curve_id должен принимать неотрицательное значение"

        curve = self.connection.get_well_curve(well_id, curve_id)
        return curve

    @staticmethod
    def parse_to_numpy(
        well_curves_list: list[datacontainers.WellCurve | datacontainers.Curve],
        min_depth: float = -np.inf,
        max_depth: float = np.inf,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Преобразовывает список кривых в numpy массив.

        Parameters
        ----------
        well_curves_list: list[cratonapi.datacontainers.WellCurve | cratonapi.datacontainers.Curve]
            Список кривых, длины N.
        min_depth: float, default=-np.inf
            Минимальная глубина.
        max_depth: float, default=np.inf
            Максимальная глубина.

        Returns
        -------
        tuple[npt.NDArray, npt.NDArray]
            Массив значений кривых, размером (M, N).
            Массив значений глубин для всех кривых, размером (M, ).
        """

        well_curves_mask = np.array(
            [well_curve.point_values is None for well_curve in well_curves_list]
        )
        depths = []
        values = []
        for well_curve in np.array(well_curves_list)[~well_curves_mask]:
            values.append(well_curve.point_values)
            depths.append(well_curve.point_depths)
        if len(values) == 0:
            raise WellEmptyCurvesException("Все заданные кривые отсутствуют")
        min_depth = np.max(
            [np.min([np.min(depths[i]) for i in range(len(depths))]), min_depth]
        )
        max_depth = np.min(
            [np.max([np.max(depths[i]) for i in range(len(depths))]), max_depth]
        )

        assert min_depth < max_depth, "Неверно заданы глубины"

        dh = np.min([np.min(np.abs(np.diff(depths[i]))) for i in range(len(depths))])
        dh = np.round(dh, 2)

        values_new = []
        full_depth = np.arange(min_depth, max_depth, dh)
        for i in range(len(depths)):
            mask = ~np.isnan(values[i])
            depths_i = depths[i][mask]
            values_i = values[i][mask]
            min_depth_i = np.max([np.min(depths_i), min_depth])
            max_depth_i = np.min([np.max(depths_i), max_depth])
            if min_depth_i >= max_depth_i:
                values_new.append(np.array([np.nan] * len(full_depth)))
                continue
            depths_i_interpolated = np.arange(min_depth_i, max_depth_i, dh)
            depths_i_interpolated = depths_i_interpolated[
                depths_i_interpolated < np.min([np.max(depths_i), max_depth])
            ]
            interp_func = interp1d(depths_i, values_i, kind="linear")
            values_i_interpolated = interp_func(depths_i_interpolated)
            idx_min = int((depths_i_interpolated[0] - full_depth[0]) // dh)
            idx_max = int((depths_i_interpolated[-1] - full_depth[0]) // dh)
            values_i_new = np.zeros(len(full_depth))
            values_i_new[:idx_min] = np.nan
            values_i_new[idx_max:] = np.nan
            values_i_new[idx_min : idx_min + len(values_i_interpolated)] = (
                values_i_interpolated
            )
            values_new.append(values_i_new)
        result_values = np.empty((len(well_curves_list), len(full_depth)))
        result_values[~well_curves_mask] = values_new
        result_values[well_curves_mask] = np.array([np.nan] * len(full_depth))
        return result_values.T, full_depth

    @staticmethod
    def parse_to_numpy_without_interpolation(
        well_curves_list: list[datacontainers.WellCurve | datacontainers.Curve],
        min_depth: float = -np.inf,
        max_depth: float = np.inf,
    ) -> tuple[list, list]:
        """
        Преобразовывает список кривых в список numpy массивов.

        Parameters
        ----------
        well_curves_list: list[cratonapi.datacontainers.WellCurve | cratonapi.datacontainers.Curve]
            Список кривых, длины N.
        min_depth: float, default=-np.inf
            Минимальная глубина.
        max_depth: float, default=np.inf
            Максимальная глубина.

        Returns
        -------
        tuple[list, list]
            Список массивов значений кривых, длины N.
            Список массивов значений глубин, длины N.
        """

        depths = []
        values = []
        for well_curve in np.array(well_curves_list):
            if well_curve.point_depths is None:
                values.append(np.array([np.nan]))
                depths.append(np.array([np.nan]))
            elif np.isnan(well_curve.point_depths).all():
                values.append(np.array([np.nan]))
                depths.append(np.array([np.nan]))
            else:
                curr_min_depth = max(np.nanmin(well_curve.point_depths), min_depth)
                curr_max_depth = min(np.nanmax(well_curve.point_depths), max_depth)

                if curr_min_depth < curr_max_depth:
                    indexes1 = np.where(well_curve.point_depths >= curr_min_depth)[0]
                    indexes2 = np.where(well_curve.point_depths <= curr_max_depth)[0]
                    indexes = [x for x in indexes1 if x in indexes2]
                    values.append(well_curve.point_values[indexes])
                    depths.append(well_curve.point_depths[indexes])
                else:
                    values.append(np.array([np.nan]))
                    depths.append(np.array([np.nan]))
        return values, depths

    @staticmethod
    def numpy_to_curve(
        values: npt.NDArray, depths: npt.NDArray, curve_name: str, curve_type: int = 0
    ) -> datacontainers.WellCurve:
        """
        Преобразовывает numpy массив в кривую.

        Parameters
        ----------
        values: npt.NDArray
            1D массив значений кривой, размером N.
        depths: npt.NDArray
            1D массив значений глубин кривой, размером N.
        curve_name: str
            Имя кривой.
        curve_type: int, default=0
            Тип кривой.

        Returns
        -------
        cratonapi.datacontainers.WellCurve
            Кривая.
        """

        assert values.ndim == 1, "values: Ожидался 1D массив"
        assert depths.ndim == 1, "depths: Ожидался 1D массив"
        assert values.shape == depths.shape, "Размеры values и depths должны быть равны"
        assert curve_type >= 0, "curve_type должен принимать неотрицательное значение"

        curve_to_save = datacontainers.WellCurve(
            curve_type=curve_type,
            curve_name=curve_name,
            point_values=values,
            point_depths=depths,
        )
        return curve_to_save

    def save(self, well_curve: datacontainers.WellCurve, well_id: int) -> None:
        """
        Загружает кривую в GISWell.

        Parameters
        ----------
        well_curve: cratonapi.datacontainers.WellCurve
            Кривая.
        well_id: int
            Идентификатор скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"

        self.connection.upload_well_curves(well_curve, well_id)

    def get_info(self) -> dict:
        """
        Подгружает информацию о скважинах.

        Returns
        -------
        dict
            Словарь, где ключ - имя скважины.
            Значения:
                id: int
                    Идентификатор скважины.
        """

        well_list = self.connection.get_wells_list()

        info = {}
        for well in well_list:
            info[well.well_name] = {"id": well.well_id}
        return info

    def get_curves_info(self, well_ids: list) -> dict:
        """
        Подгружает информацию о кривых в скважинах.

        Parameters
        ----------
        well_ids: list
            Идентификаторы скважин.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор скважины.
            Значения:
                curve_info: dict
                    Словарь, где ключ - имя кривой.
                        Значения:
                            id: int
                                Идентификатор кривой.
                            type: int
                                Тип(тэг) кривой.
                            visibility: int
                                Видимость кривой.
                            start_depth: float
                                Глубина начала кривой.
                            end_depth: float
                                Глубина конца кривой.
                            dh: float
                                Шаг глубины кривой.
                            min_value: float
                                Минимальное значение кривой.
                            max_value: float
                                Максимальное  значение кривой.
                            intervals: ndarray
                                Интервалы существования кривой.
        """

        info = {}
        for well_id in well_ids:
            info[well_id] = {"curve_info": self.get_curves_info_dict(well_id)}
        return info

    def get_strat_levels_info(
        self, well_ids: list, all_strat_levels_info: dict
    ) -> dict:
        """
        Подгружает информацию о стратиграфических уровнях в скважинах.

        Parameters
        ----------
        well_ids: list
            Идентификаторы скважин.
        all_strat_levels_info: dict
            Словарь, где ключ - идентификатор стратиграфического уровня.
            Значения:
                'level_age': float
                    Возраст уровня (млн.лет).
                'level_name': str
                    Имя стратиграфического уровня.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор скважины.
            Значения:
                strat_levels: dict
                    Словарь, где ключ - имя стратиграфического уровня.
                        Значения:
                            'level_id': int
                                Идентификатор стратиграфического уровня.
                            'level_depth': int
                                Глубина стратиграфического уровня.
                            'level_age': float
                                Возраст уровня (млн.лет).
        """

        info = {}
        for well_id in well_ids:
            well_levels_depth_dict = self.get_well_levels_depths(well_id)
            well_strat_info = Well.__parse_well_dicts(
                well_levels_depth_dict, all_strat_levels_info
            )
            info[well_id] = {"strat_levels": well_strat_info}
        return info

    def get_curves_info_dict(self, well_id: int) -> dict:
        """
        Подгружает информацию о кривых в скважине.

        Returns
        -------
        dict
            Словарь, где ключ - имя кривой.
            Значения:
                id: int
                    Идентификатор кривой.
                type: int
                    Тип кривой.
                visibility: int
                    Видимость кривой.
                start_depth: float
                    Глубина начала кривой.
                end_depth: float
                    Глубина конца кривой.
                dh: float
                    Шаг глубины кривой.
                min_value: float
                    Минимальное значение кривой.
                max_value: float
                    Максимальное  значение кривой.
                intervals: ndarray
                    Интервалы существования кривой.
        """

        curves_list = self.connection.get_curves_list(well_id)
        info = {}
        for curve in curves_list:
            id_dict = {"id": curve.curve_id}
            type_dict = {"type": curve.curve_type}
            visibility_dict = {"visibility": curve.curve_visibility}
            start_depth = {"start_depth": curve.start_depth}
            end_depth = {"end_depth": curve.end_depth}
            dh = {"dh": curve.dh}
            min_value = {"min_value": curve.min_value}
            max_value = {"max_value": curve.max_value}
            intervals = {"intervals": curve.intervals}

            info_dict = dict(
                **id_dict,
                **type_dict,
                **visibility_dict,
                **start_depth,
                **end_depth,
                **dh,
                **min_value,
                **max_value,
                **intervals
            )
            info[curve.curve_name] = info_dict
        return info

    def get_curve_display_properties(self) -> dict:
        """
        Подгружает информацию об отображении кривых.

        Returns
        -------
        dict
            Словарь, где ключ - имя тэга кривой.
            Значения:
                id: int
                    Идентификатор тэга кривой.
                description: str
                    Описание тэга.
                priority: int
                    Приоритет тэга.
                type_interpolation: int
                    Тип интерполяции 0 — гладкая, 1 - ступенчатая.
                type_display: int
                    Тип отображения 0 — ломаные линии, 1 - точки.
                type_scale: int
                    Тип масштаба 0 — линейный, 1 - логарифмический.
                auto_scaling: int
                    Автоматический подбор масштаба
                manual_scaling_interval: ndarray
                    Начало и конец для ручного масштаба.
                manual_scaling_step: int
                    Кратность для ручного масштаба.
                line_width: float
                    Толщина линии.
                line_color: tuple[float, float, float, float]
                    Цвет отображения кривой в GISWell(Красный, Зелёный, Синий, Альфа). Значения от 0 до 1.
                filling: int
                    Заливка.
                filling_direction: int
                    Направление заливки.
                filling_color1: Color
                    Первый цвет заливки.
                filling_color2: Color
                    Второй цвет заливки.
                filling_interval: ndarray
                    Процент начала и конца заливки.
        """

        curve_display_properties = self.connection.get_curve_display_properties()
        info = {}
        for tag in curve_display_properties:
            id_dict = {"id": tag.tag_id}
            description_dict = {"description": tag.description}
            tag_priority_dict = {"priority": tag.tag_priority}
            type_interpolation_dict = {"type_interpolation": tag.type_interpolation}
            type_display_dict = {"type_display": tag.type_display}
            type_scale_dict = {"type_scale": tag.type_scale}
            auto_scaling_dict = {"auto_scaling": tag.auto_scaling}
            manual_scaling_interval_dict = {
                "manual_scaling_interval": tag.manual_scaling_interval
            }
            manual_scaling_step_dict = {"manual_scaling_step": tag.manual_scaling_step}
            line_width_dict = {"line_width": tag.line_width}
            line_color_dict = {
                "line_color": (
                    tag.line_color.red / 255,
                    tag.line_color.green / 255,
                    tag.line_color.blue / 255,
                    tag.line_color.alpha / 255,
                )
            }
            filling_dict = {"filling": tag.filling}
            filling_direction_dict = {"filling_direction": tag.filling_direction}
            filling_color1_dict = {"filling_color1": tag.filling_color1}
            filling_color2_dict = {"filling_color2": tag.filling_color2}
            filling_interval_dict = {"filling_interval": tag.filling_interval}

            info_dict = dict(
                **id_dict,
                **description_dict,
                **tag_priority_dict,
                **type_interpolation_dict,
                **type_display_dict,
                **type_scale_dict,
                **auto_scaling_dict,
                **manual_scaling_interval_dict,
                **manual_scaling_step_dict,
                **line_width_dict,
                **line_color_dict,
                **filling_dict,
                **filling_direction_dict,
                **filling_color1_dict,
                **filling_color2_dict,
                **filling_interval_dict
            )
            info[tag.tag_name] = info_dict
        return info

    def get_all_strat_levels_dict(self) -> dict:
        """
        Подгружает информацию о всех стратиграфических уровнях.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор стратиграфического уровня.
            Значения:
                'level_age': float
                    Возраст уровня (млн.лет).
                'level_name': str
                    Имя стратиграфического уровня.
        """

        strat_levels = self.connection.get_stratigraphic_levels()
        strat_levels_dict = {}
        for i in range(len(strat_levels)):
            strat_levels_dict[strat_levels[i].level_id] = {
                "level_age": strat_levels[i].level_age,
                "level_name": strat_levels[i].level_name,
            }
        return strat_levels_dict

    def get_well_levels_depths(self, well_id: int) -> dict:
        """
        Подгружает глубины стратиграфических уровней в скважине.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор стратиграфического уровня.
            Значения:
                'level_depth': int
                    Глубина стратиграфического уровня.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"

        well_levels_depth = self.connection.get_well_stratigraphic_levels(well_id)
        well_levels_depth_dict = {}
        for i in range(len(well_levels_depth)):
            well_levels_depth_dict[well_levels_depth[i].level_id] = {
                "level_depth": well_levels_depth[i].level_depth
            }
        return well_levels_depth_dict

    @staticmethod
    def __parse_well_dicts(
        well_levels_depth_dict: dict, strat_levels_dict: dict
    ) -> dict:
        """Парсит словари полученные из функций get_all_strat_levels_dict и get_well_levels_depths."""

        well_strat_info = {}
        for level_id in well_levels_depth_dict.keys():
            level_name = strat_levels_dict[level_id]["level_name"]
            well_strat_info[level_name] = {
                "level_id": level_id,
                "level_depth": well_levels_depth_dict[level_id]["level_depth"],
                "level_age": strat_levels_dict[level_id]["level_age"],
            }
        return well_strat_info

    @staticmethod
    def get_min_max_depth(
        well_curves_list: list[datacontainers.WellCurve],
    ) -> tuple[float, float]:
        """
        Считает минимальную и максимальную глубину среди всех кривых.

        Parameters
        ----------
        well_curves_list: list[cratonapi.datacontainers.WellCurve]
            Список кривых.

        Returns
        -------
        tuple[float, float]
            Минимальная глубина среди всех кривых.
            Максимальная глубина среди всех кривых.
        """

        depths = [curve.point_depths for curve in well_curves_list]
        min_depth = np.min([np.min(depths[i]) for i in range(len(depths))])
        max_depth = np.max([np.max(depths[i]) for i in range(len(depths))])
        return min_depth, max_depth

    @staticmethod
    def __get_coord(z: float, z0: float, z1: float, x0: float, x1: float) -> float:
        return ((z - z0) / (z1 - z0)) * (x1 - x0) + x0

    @staticmethod
    def get_strat_depth(
        strat_dict: dict, strat_info: dict, else_number: float = np.nan
    ) -> float:
        """
        Считает глубину отбивки.

        Parameters
        ----------
        strat_dict: dict
            Словарь со значениями:
                'name': str
                    Имя стратиграфического уровня.
                'shift': int | float
                    Отступ от стратиграфического уровня в метрах.
                'direction': int
                    Направление отступа от стратиграфического уровня: 1 - вниз, -1 - вверх.
        strat_info: dict
            Словарь, где ключ - имя стратиграфического уровня.
                Значение:
                    Словарь со значением:
                        'level_depth': int
                            Глубина стратиграфического уровня.
        else_number: float, default=np.nan
            Значение глубины, если стратиграфического уровня нет.

        Returns
        -------
        float
            Глубина отбивки.
        """
        if strat_dict["name"] in list(strat_info.keys()):
            depth = strat_info[strat_dict["name"]]["level_depth"]
            depth = depth + strat_dict["shift"] * strat_dict["direction"]
        else:
            depth = else_number
        return depth

    @staticmethod
    def __get_distance(coords_a: npt.NDArray, coords_b: npt.NDArray) -> float:
        """Вычисляет расстояние между точками."""

        return np.sqrt(
            (coords_a[0] - coords_b[0]) ** 2 + (coords_a[1] - coords_b[1]) ** 2
        )

    @staticmethod
    def __get_well_coords_by_depth(
        trajectory_coords: npt.NDArray[np.float_],
        well_coords_initial: npt.NDArray[np.float_],
        depth: float,
    ) -> Tuple[float, float]:
        if np.isnan(depth):
            x = np.nan
            y = np.nan
        else:
            z0 = float(trajectory_coords[0, 2])
            z1 = float(trajectory_coords[1, 2])
            x0 = float(trajectory_coords[0, 0])
            x1 = float(trajectory_coords[1, 0])
            y0 = float(trajectory_coords[0, 1])
            y1 = float(trajectory_coords[1, 1])

            x = Well.__get_coord(depth, z0, z1, x0, x1)
            x = well_coords_initial[0] + x
            y = Well.__get_coord(depth, z0, z1, y0, y1)
            y = well_coords_initial[1] + y
        return x, y

    @staticmethod
    def __get_depth_idx(well_trajectory: npt.NDArray[np.float_], depth: float) -> int:
        if np.isnan(depth):
            idx = 0
        else:
            n = len(well_trajectory) - 1
            tolerance = np.zeros(n)
            for i in range(n):
                tolerance[i] = well_trajectory[i + 1, 2] - depth
            condition = tolerance < 0
            tolerance[condition] = np.nan
            idx = int(np.nanargmin(tolerance) + 1)
        return idx

    def get_well_coords_by_trajectory(
        self,
        strat_info: dict,
        well_id: int,
        strat_levels_parameters: dict,
        coords_calculation_parameters: dict,
        well_name: Optional[str] = None,
    ) -> dict:
        """
         Считает координаты кровли, подошвы и середины.

         Parameters
         ----------
         strat_info: dict
             Словарь, где ключ - имя стратиграфического уровня.
                 Значение:
                     Словарь со значением:
                         'level_depth': int
                             Глубина стратиграфического уровня.
         well_id: int
             Идентификатор скважины.
         well_name: Optional[str]
             Имя скважины.
         strat_levels_parameters: dict
             Словарь, со значениями:
                 'top': dict
                     Верхняя отбивка(кровля). Словарь, со значениями:
                         'name': str
                             Имя стратиграфического уровня.
                         'shift': Union[int, float]
                             Отступ от стратиграфического уровня в метрах.
                         'direction': int
                             Направление отступа от стратиграфического уровня: 1 - вниз, -1 - вверх.
                 'bot': dict
                     Нижняя отбивка(подошва). Словарь, со значениями:
                         'name': str
                             Имя стратиграфического уровня.
                         'shift': Union[int, float]
                             Отступ от стратиграфического уровня в метрах.
                         'direction': int
                             Направление отступа от стратиграфического уровня: 1 - вниз, -1 - вверх.
         coords_calculation_parameters: dict
             Словарь, со значениями:
                 'depth_type': str
                     Тип глубины: 'VD' - вертикальная, 'MD' - измеренная.
                 'distance': Union[int, float]
                     Порог отброса по горизонтальному расстоянию (метры).

         Returns
         -------
        dict
             Словарь, со значениями:
                 'state': bool
                     Состояние, обозначающее можно ли использовать скважину дальше в расчетах или нет.(True - используется, False - не используется)
                 'top': npt.NDArray
                     Массив [x, y] координат кровли.
                 'bot': npt.NDArray
                     Массив [x, y] координат подошвы.
                 'middle': npt.NDArray
                     Массив [x, y] координат середины.
        """

        top_dict = strat_levels_parameters["top"]
        bot_dict = strat_levels_parameters["bot"]
        depth_type = coords_calculation_parameters["depth_type"]
        distance = coords_calculation_parameters["distance"]

        depth_top = Well.get_strat_depth(top_dict, strat_info)
        depth_bot = Well.get_strat_depth(bot_dict, strat_info)
        depth_middle = float(np.mean([depth_top, depth_bot]))

        well_metadata = self.connection.get_wells_data(1, well_id)[0]
        well_coords_initial = np.array(
            [well_metadata.outfall_x, well_metadata.outfall_y]
        )
        well_trajectory_cls = self.connection.get_well_trajectory(well_id)

        if depth_type == "VD":
            well_trajectory = np.array(
                [
                    well_trajectory_cls.point_x_shifts,
                    well_trajectory_cls.point_y_shifts,
                    well_trajectory_cls.point_depths,
                ]
            ).T
        elif depth_type == "MD":
            well_trajectory = np.array(
                [
                    well_trajectory_cls.point_x_shifts,
                    well_trajectory_cls.point_y_shifts,
                    well_trajectory_cls.point_z_shifts,
                ]
            ).T
        else:
            raise RuntimeError(
                "Задан неправильный тип глубины(depth_type). Допустимые типы глубин: 'VD' - вертикальная, 'MD' - измеренная."
            )

        if np.nanmax(well_trajectory[:, 2]) < depth_top:
            raise WellTrajectoryDepthException(
                "Для скважины {well_name} отбивка {strat_name} лежит ниже максимальной {depth_type} глубины".format(
                    well_name=well_name,
                    strat_name=top_dict["name"],
                    depth_type="вертикальной" if depth_type == "VD" else "измеренной",
                )
            )

        if np.nanmax(well_trajectory[:, 2]) < depth_bot:
            raise WellTrajectoryDepthException(
                "Для скважины {well_name} отбивка {strat_name} лежит ниже максимальной {depth_type} глубины".format(
                    well_name=well_name,
                    strat_name=bot_dict["name"],
                    depth_type="вертикальной" if depth_type == "VD" else "измеренной",
                )
            )

        top_idx = Well.__get_depth_idx(well_trajectory, depth_top)
        bot_idx = Well.__get_depth_idx(well_trajectory, depth_bot)
        middle_idx = Well.__get_depth_idx(well_trajectory, depth_middle)

        (top_x, top_y) = Well.__get_well_coords_by_depth(
            well_trajectory[top_idx - 1 : top_idx + 1], well_coords_initial, depth_top
        )

        (bot_x, bot_y) = Well.__get_well_coords_by_depth(
            well_trajectory[bot_idx - 1 : bot_idx + 1], well_coords_initial, depth_bot
        )

        (middle_x, middle_y) = Well.__get_well_coords_by_depth(
            well_trajectory[middle_idx - 1 : middle_idx + 1],
            well_coords_initial,
            depth_middle,
        )

        calculated_distance = Well.__get_distance(
            np.array([top_x, top_y]), np.array([bot_x, bot_y])
        )
        state = not (
            (calculated_distance > float(distance)) or np.isnan(calculated_distance)
        )

        coords_dict = {
            "state": state,
            "top": np.array([top_x, top_y]),
            "bot": np.array([bot_x, bot_y]),
            "middle": np.array([middle_x, middle_y]),
        }
        return coords_dict



import numpy as np
from scipy import stats
from sklearn.cluster import HDBSCAN, KMeans
from sklearn.mixture import GaussianMixture
from typing_extensions import Optional

RANDOM_STATE = 42
ALPHA_K = 0.02


class KMeansClassifier:
    """Класс для KMeans кластеризации."""

    @staticmethod
    def __get_scaled_inertia(data: np.ndarray, cluster: KMeans) -> np.ndarray:
        """Считает взвешенную инерцию."""

        inertia_o = np.square((data - data.mean(axis=0))).sum()
        scaled_inertia = cluster.inertia_ / inertia_o + ALPHA_K * cluster.n_clusters
        return scaled_inertia

    @staticmethod
    def get_cluster(data: np.ndarray, n_clusters: int) -> KMeans:
        cluster = KMeans(
            n_clusters=n_clusters, random_state=RANDOM_STATE, algorithm="elkan"
        )
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_cluster_centers(kmeans: KMeans) -> np.ndarray:
        kmeans_centers = np.unique(np.sum(kmeans.cluster_centers_, axis=1))
        return kmeans_centers

    @staticmethod
    def __get_labels(kmeans: KMeans) -> np.ndarray:
        """Возвращает список меток."""

        kmeans_labels_before_sorting = kmeans.labels_
        kmeans_centers = KMeansClassifier.__get_cluster_centers(kmeans)
        kmeans_uniq_labels = np.unique(kmeans_labels_before_sorting)
        sorted_kmeans_uniq_labels = kmeans_uniq_labels[np.argsort(kmeans_centers)]

        kmeans_labels = np.zeros_like(kmeans_labels_before_sorting)
        for i, label in enumerate(sorted_kmeans_uniq_labels):
            kmeans_labels[kmeans_labels_before_sorting == label] = i

        return kmeans_labels

    @staticmethod
    def find_the_best_number_of_clusters(
        data: np.ndarray,
    ) -> tuple[range, np.ndarray, int]:
        """
        Подбирает оптимальное количество кластеров на основании взвешенной инерции.

        Parameters
        ----------
        data: np.ndarray
            2D массив значений.

        Returns
        -------
        tuple[range, np.ndarray, int]
            Список количества кластеров для которых производился подсчет.
            Список взвешенных инерций для каждого количества кластеров.
            Оптимальное количество кластеров.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"

        scaled_inertia_list = []
        number_of_clusters_list = range(2, min(15, data.shape[0]))
        for number_of_clusters in number_of_clusters_list:
            cluster = KMeansClassifier.get_cluster(data, number_of_clusters)
            scaled_inertia = KMeansClassifier.__get_scaled_inertia(data, cluster)
            scaled_inertia_list.append(scaled_inertia)
        scaled_inertia_array = np.asarray(scaled_inertia_list)
        best_number_of_clusters = number_of_clusters_list[
            np.argmin(scaled_inertia_array)
        ]
        return number_of_clusters_list, scaled_inertia_array, best_number_of_clusters

    @staticmethod
    def calculate(data: np.ndarray, n_clusters: int) -> tuple[KMeans, np.ndarray]:
        """
        Распределяет объекты по n_clusters кластерам по сходству. Для вычисления этого сходства используется евклидово расстояние в качестве меры.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N-количество точек, M - количество атрибутов.
        n_clusters: int
            Количество кластеров.

        Returns
        -------
        tuple[KMeans, np.ndarray]
            Объект класса KMeans.
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        assert (
            n_clusters > 0
        ), "Параметр n_clusters должен принимать положительное значение"
        cluster = KMeansClassifier.get_cluster(data, n_clusters)
        labels = KMeansClassifier.__get_labels(cluster)
        return cluster, labels


class GaussianMixtureClassifier:
    """Класс для Gaussian Mixture кластеризации."""

    @staticmethod
    def get_cluster(data: np.ndarray, n_clusters: int) -> GaussianMixture:
        cluster = GaussianMixture(n_components=n_clusters, random_state=RANDOM_STATE)
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_cluster_centers(gm: GaussianMixture, data: np.ndarray) -> np.ndarray:
        centers = np.empty(shape=(gm.n_components, data.shape[1]))
        for i in range(gm.n_components):
            density = stats.multivariate_normal(
                cov=gm.covariances_[i], mean=gm.means_[i]
            ).logpdf(data)
            centers[i, :] = data[np.argmax(density)]
        centers = np.sum(centers, axis=1)
        return centers

    @staticmethod
    def __get_labels(gm: GaussianMixture, data: np.ndarray) -> np.ndarray:
        """Возвращает массив меток."""

        gm_labels_before_sorting = gm.predict(data)
        gm_centers = GaussianMixtureClassifier.__get_cluster_centers(gm, data)
        gm_uniq_labels = np.unique(gm_labels_before_sorting)
        sorted_gm_uniq_labels = gm_uniq_labels[np.argsort(gm_centers)]

        gm_labels = np.zeros_like(gm_labels_before_sorting)
        for i, label in enumerate(sorted_gm_uniq_labels):
            gm_labels[gm_labels_before_sorting == label] = i

        return gm_labels

    @staticmethod
    def find_the_best_number_of_clusters(
        data: np.ndarray,
    ) -> tuple[range, np.ndarray, int]:
        """
        Подбирает оптимальное количество кластеров на основе Байесовского критерия.

        Parameters
        ----------
        data: np.ndarray
            2D массив значений.

        Returns
        -------
        tuple[range, np.ndarray, int]
            Список количества кластеров для которых производился подсчет.
            Список Байесовских критериев для каждого количества кластеров.
            Оптимальное количество кластеров.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"

        bic_list = []
        number_of_clusters_list = range(2, min(15, data.shape[0]))
        for number_of_clusters in number_of_clusters_list:
            gm = GaussianMixtureClassifier.get_cluster(data, number_of_clusters)
            bic = -gm.bic(data)
            if bic < 0:
                bic = np.nan
            bic_list.append(bic)
        bic_array = np.asarray(bic_list)
        # bic_list_log = np.log(bic_list)
        # best_number_of_clusters = number_of_clusters_list[np.nanargmin(bic_list_log)]
        best_number_of_clusters = number_of_clusters_list[np.nanargmin(bic_array)]
        return number_of_clusters_list, bic_array, best_number_of_clusters

    @staticmethod
    def calculate(
        data: np.ndarray, n_clusters: int
    ) -> tuple[GaussianMixture, np.ndarray]:
        """
        Распределяет объекты по n_clusters кластерам, предполагая, что данные состоят из смеси гауссовых распределений.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N - количество точек, M - количество атрибутов.
        n_clusters: int
            Количество кластеров.

        Returns
        -------
        tuple[GaussianMixture, np.ndarray]
            Объект класса GaussianMixture.
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        assert (
            n_clusters > 0
        ), "Параметр n_clusters должен принимать положительное значение"
        cluster = GaussianMixtureClassifier.get_cluster(
            data=data, n_clusters=n_clusters
        )
        labels = GaussianMixtureClassifier.__get_labels(gm=cluster, data=data)
        return cluster, labels


class HDBSCANClassifier:
    """Класс для HDBSCAN кластеризации."""

    @staticmethod
    def get_cluster(
        data: np.ndarray,
        min_cluster_size: Optional[int],
        min_samples: Optional[int],
        cluster_selection_epsilon: float,
    ) -> HDBSCAN:
        cluster = HDBSCAN(
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            cluster_selection_epsilon=cluster_selection_epsilon,
        )
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_labels(hdbscan: HDBSCAN) -> np.ndarray:
        """Возвращает список меток."""

        hdbscan_labels = hdbscan.labels_
        return hdbscan_labels

    @staticmethod
    def get_number_of_clusters(labels: np.ndarray) -> int:
        """
        Возвращает количество кластеров.

        Parameters
        ----------
        labels: np.ndarray
            1D массив меток.

        Returns
        -------
        int
            Количество кластеров.
        """

        assert labels.ndim == 1, "labels: Ожидался 1D массив"

        return np.unique(labels).shape[0]

    @staticmethod
    def calculate(
        data: np.ndarray,
        min_cluster_size: int = 5,
        min_samples: Optional[int] = None,
        cluster_selection_epsilon: float = 0.0,
    ) -> tuple[HDBSCAN, np.ndarray]:
        """
        Выявляет кластеры в наборе данных на основе распределения плотности точек данных.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N-количество точек, M - количество атрибутов.
        min_cluster_size: int, default=5
            Минимальное количество выборок, чтобы эта группа считалась кластером(параметр HDBSCAN).
        min_samples: int, default=None
            Минимальное количество выборок в окрестности, чтобы точка считалась центральной(параметр HDBSCAN).
        cluster_selection_epsilon: float, default=0.0
            Максимальное расстояние, допустимое между точками, чтобы они считались связанными в процессе кластеризации на основе плотности(параметр HDBSCAN).

        Returns
        -------
        tuple[HDBSCAN, np.ndarray]
            Объект класса HDBSCAN.
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        cluster = HDBSCANClassifier.get_cluster(
            data, min_cluster_size, min_samples, cluster_selection_epsilon
        )
        labels = HDBSCANClassifier.__get_labels(cluster)
        return cluster, labels
